#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "egbbdll.h"
#include "common.h"

static const int WIN_SCORE  =  5000;

enum {
	DECOMP_IN_RAM,DECOMP_IN_DISK,COMP_IN_RAM,COMP_IN_DISK
};
static const int VALUE[4] = {
	DRAW, WIN, LOSS, PREDICTED
};

#define is_in_disk(x)    ((x) & 1)
#define is_comp(x)       ((x) & 2)

static SEARCHER searchers[8];
static LOCK searcher_lock;

int COMPT[0xf3] = {
0x000, 0x001, 0x002, 0x004, 0x005, 0x006, 0x008, 0x009, 0x00a, 0x010, 0x011, 0x012, 0x014, 0x015, 0x016,
0x018, 0x019, 0x01a, 0x020, 0x021, 0x022, 0x024, 0x025, 0x026, 0x028, 0x029, 0x02a, 0x040, 0x041, 0x042,
0x044, 0x045, 0x046, 0x048, 0x049, 0x04a, 0x050, 0x051, 0x052, 0x054, 0x055, 0x056, 0x058, 0x059, 0x05a,
0x060, 0x061, 0x062, 0x064, 0x065, 0x066, 0x068, 0x069, 0x06a, 0x080, 0x081, 0x082, 0x084, 0x085, 0x086,
0x088, 0x089, 0x08a, 0x090, 0x091, 0x092, 0x094, 0x095, 0x096, 0x098, 0x099, 0x09a, 0x0a0, 0x0a1, 0x0a2,
0x0a4, 0x0a5, 0x0a6, 0x0a8, 0x0a9, 0x0aa, 0x100, 0x101, 0x102, 0x104, 0x105, 0x106, 0x108, 0x109, 0x10a,
0x110, 0x111, 0x112, 0x114, 0x115, 0x116, 0x118, 0x119, 0x11a, 0x120, 0x121, 0x122, 0x124, 0x125, 0x126,
0x128, 0x129, 0x12a, 0x140, 0x141, 0x142, 0x144, 0x145, 0x146, 0x148, 0x149, 0x14a, 0x150, 0x151, 0x152,
0x154, 0x155, 0x156, 0x158, 0x159, 0x15a, 0x160, 0x161, 0x162, 0x164, 0x165, 0x166, 0x168, 0x169, 0x16a,
0x180, 0x181, 0x182, 0x184, 0x185, 0x186, 0x188, 0x189, 0x18a, 0x190, 0x191, 0x192, 0x194, 0x195, 0x196,
0x198, 0x199, 0x19a, 0x1a0, 0x1a1, 0x1a2, 0x1a4, 0x1a5, 0x1a6, 0x1a8, 0x1a9, 0x1aa, 0x200, 0x201, 0x202,
0x204, 0x205, 0x206, 0x208, 0x209, 0x20a, 0x210, 0x211, 0x212, 0x214, 0x215, 0x216, 0x218, 0x219, 0x21a,
0x220, 0x221, 0x222, 0x224, 0x225, 0x226, 0x228, 0x229, 0x22a, 0x240, 0x241, 0x242, 0x244, 0x245, 0x246,
0x248, 0x249, 0x24a, 0x250, 0x251, 0x252, 0x254, 0x255, 0x256, 0x258, 0x259, 0x25a, 0x260, 0x261, 0x262,
0x264, 0x265, 0x266, 0x268, 0x269, 0x26a, 0x280, 0x281, 0x282, 0x284, 0x285, 0x286, 0x288, 0x289, 0x28a,
0x290, 0x291, 0x292, 0x294, 0x295, 0x296, 0x298, 0x299, 0x29a, 0x2a0, 0x2a1, 0x2a2, 0x2a4, 0x2a5, 0x2a6,
0x2a8, 0x2a9, 0x2aa
};

/*
EGBB
*/
EGBB* egbbs[TOTAL_EGBB_COUNT];
static const char piece_name[] = "_KQRBNPkqrbnp_";

int EGBB::GetIndex(int side,int piece1,int piece2,int piece3) {
	ENUMERATOR myenum;
	myenum.player = side;
	myenum.add(wking);
	myenum.add(bking);
	myenum.add(piece1);
	if(piece2) myenum.add(piece2);
	if(piece3) myenum.add(piece3);
	return GetIndex(&myenum);
}
int EGBB::GetIndex(ENUMERATOR* penum) {
	register int index = penum->player,prd = 2;
	penum->sort();
	for(int i = 2;i < penum->n_piece;i++) {
		index += penum->piece[i] * prd;
		prd *= 14;
	}
	return index;
}
/*
Open EGBB
*/
void EGBB::open(char* path,int egbb_state) {
    UBMP32 i;
	
    //load file according to egbb_state	
	state = egbb_state;
	is_loaded = false;

	strcpy(name,path);
	strcat(name,enumerator.name);
	if(is_comp(state)) {
		strcat(name,".cmp");
	} else {
		strcat(name,".bin");
	}
		
	//Decompresed in ram/disk
	if(!is_comp(state)) {  
		
		pf = fopen(path,"rb");
		if(!pf)
			return;
		
		//get size of file and allocate buffers
		UBMP32 TB_SIZE;
		fseek(pf,0,SEEK_END);
		TB_SIZE = ftell(pf);
		fseek(pf,0,SEEK_SET);
		
		//load in ram
		if(state == DECOMP_IN_RAM) {
			table = new UBMP8[TB_SIZE];
			for(i = 0;i < TB_SIZE;i++)
				table[i] = fgetc(pf);
		}
	//compresed in ram/disk
	} else {              
		if(!COMP_INFO::open(name))
			return;

		//compressed files in RAM
		if(state == COMP_IN_RAM) {
			table = new UBMP8[cmpsize];
			
			for(i = 0;i < cmpsize;i++) {
				table[i] = fgetc(pf);
			}
		}
	}

	is_loaded = true;
}

DLLExport void load_egbb_into_ram(int side,int piece1,int piece2,int piece3) {
	EGBB* pegbb = egbbs[EGBB::GetIndex(side,piece1,piece2,piece3)];
	if(pegbb->state != COMP_IN_RAM) {
		pegbb->table = new UBMP8[pegbb->cmpsize];
		for(UBMP32 i = 0;i < pegbb->cmpsize;i++) {
			pegbb->table[i] = fgetc(pegbb->pf);
		}
		pegbb->state = COMP_IN_RAM;
	}
}

DLLExport void unload_egbb_from_ram(int side,int piece1,int piece2,int piece3) {
    EGBB* pegbb = egbbs[EGBB::GetIndex(side,piece1,piece2,piece3)];
    if(pegbb->state == COMP_IN_RAM) {
		pegbb->state = COMP_IN_DISK;
		delete[] (pegbb->table);
	}
}
/*
get score of indexed position
*/

int EGBB::get_score(UBMP32 index,PSEARCHER psearcher) {
	int score,q,r;
	UBMP8 value;

	q = (index / 4);
	r = (index % 4);

	if(state == DECOMP_IN_RAM) {
        value = table[q];
	} else if(state == DECOMP_IN_DISK) {
		UBMP32 block_start = (q / BLOCK_SIZE) * BLOCK_SIZE;
		UBMP32 probe_index = q - block_start;
		
		psearcher->info.start_index = block_start;
		
        if(LRUcache.get(psearcher->info.start_index,probe_index,value) == CACHE_HIT) {
		} else {
			l_lock(lock);
			fseek(pf,block_start,SEEK_SET);
			fread(&psearcher->info.block,BLOCK_SIZE,1,pf);
			l_unlock(lock);
			value = psearcher->info.block[probe_index];
            LRUcache.add(&psearcher->info);
		}
	} else if(state == COMP_IN_RAM) {
		UBMP32 block_size;
		UBMP32 n_blk = (q / BLOCK_SIZE);
		UBMP32 probe_index = q - (n_blk * BLOCK_SIZE);
		
		psearcher->info.start_index = index_table[n_blk];
		
		if(LRUcache.get(psearcher->info.start_index,probe_index,value) == CACHE_HIT) {
		} else {
			block_size = index_table[n_blk + 1] - index_table[n_blk];
			block_size = decode(&table[psearcher->info.start_index],psearcher->info.block,block_size);
			value = psearcher->info.block[probe_index];
			LRUcache.add(&psearcher->info);
		}
	}  else if(state == COMP_IN_DISK) {
		UBMP32 block_size;
		UBMP32 n_blk = (q / BLOCK_SIZE);
		UBMP32 probe_index = q - (n_blk * BLOCK_SIZE);

		psearcher->info.start_index = index_table[n_blk];
		
		if(LRUcache.get(psearcher->info.start_index,probe_index,value) == CACHE_HIT) {
		} else {
			
			block_size = index_table[n_blk + 1] - index_table[n_blk];

			l_lock(lock);
			fseek(pf,read_start + psearcher->info.start_index,SEEK_SET);
			fread(psearcher->temp_block,block_size,1,pf);
			l_unlock(lock);

        	block_size = decode(psearcher->temp_block,psearcher->info.block,block_size);

			value = psearcher->info.block[probe_index];
			LRUcache.add(&psearcher->info);
		}
	}

	score = VALUE[(value >> (r << 1)) & 3];

    return score;
}
/*
check presences
*/
static void check_presence(EGBB** pegbb) {
	if(pegbb[0]->is_loaded) {
		if(!pegbb[1]->is_loaded) {
			pegbb[1]->use_search = true;
		}
	} else if(pegbb[1]->is_loaded) {
		pegbb[0]->use_search = true;
	}
}
/*
Open EGBB files and allocate cache
*/
void load_egbb_xxx(char* path,int cache_size,int load_options) {
	EGBB* pegbb[2];
	ENUMERATOR* penum;
	int side,piece1,piece2,piece3,state1,state2,tab_index;

	if(load_options == LOAD_NONE) {
        state1 = COMP_IN_DISK;
		state2 = COMP_IN_DISK;
	} else if(load_options == LOAD_4MEN) {
		state1 = COMP_IN_RAM;
		state2 = COMP_IN_DISK;
	} else if(load_options == SMART_LOAD) {
		state1 = COMP_IN_RAM;
		state2 = COMP_IN_DISK;
	} else if(load_options == LOAD_5MEN) {
		state1 = COMP_IN_RAM;
		state2 = COMP_IN_RAM;
	}

	printf("EgbbProbe 3.3 by Daniel Shawul\n");
	fflush(stdout);

	init_sqatt();
	init_indices();
	LRU_CACHE::alloc( cache_size );
    l_create(searcher_lock);
	for(int i = 0;i < TOTAL_EGBB_COUNT;i++) 
		egbbs[i] = 0;

	printf("Loading egbbs....");
	fflush(stdout);

	for(piece1 = wqueen; piece1 <= wpawn; piece1++) {
		for(side = white; side <= black;side++) {
            tab_index = EGBB::GetIndex(side,piece1);
			pegbb[side] = egbbs[tab_index] = new EGBB;
			penum = &pegbb[side]->enumerator;
			penum->player = side;
			penum->add(wking);
			penum->add(piece1);
			penum->add(bking);
			penum->init();
			pegbb[side]->open(path,state1);
		}
		check_presence(pegbb);
	}
	for(piece1 = wqueen; piece1 <= wpawn; piece1++) {
		for(piece2 = piece1; piece2 <= wpawn; piece2++) {
			for(side = white; side <= black;side++) {
				tab_index = EGBB::GetIndex(side,piece1,piece2);
			    pegbb[side] = egbbs[tab_index] = new EGBB;
				penum = &pegbb[side]->enumerator;
				penum->player = side;
				penum->add(wking);
				penum->add(piece1);
				penum->add(piece2);
				penum->add(bking);
				penum->init();
				pegbb[side]->open(path,state1);
			}
			check_presence(pegbb);
		}
	}
	for(piece1 = wqueen; piece1 <= wpawn; piece1++) {
		for(piece2 = COMBINE(black,PIECE(piece1)); piece2 <= bpawn; piece2++) {
			for(side = white; side <= black;side++) {
				tab_index = EGBB::GetIndex(side,piece1,piece2);
			    pegbb[side] = egbbs[tab_index] = new EGBB;
				penum = &pegbb[side]->enumerator;
				penum->player = side;
				penum->add(wking);
				penum->add(piece1);
				penum->add(bking);
				penum->add(piece2);
				penum->init();
				pegbb[side]->open(path,state1);
			}
			check_presence(pegbb);
		}
	}
	for(piece1 = wqueen; piece1 <= wpawn; piece1++) {
		for(piece2 = piece1; piece2 <= wpawn; piece2++) {
			for(piece3 = piece2; piece3 <= wpawn; piece3++) {
				for(side = white; side <= black;side++) {
					tab_index = EGBB::GetIndex(side,piece1,piece2,piece3);
			        pegbb[side] = egbbs[tab_index] = new EGBB;
					penum = &pegbb[side]->enumerator;
					penum->player = side;
					penum->add(wking);
					penum->add(piece1);
					penum->add(piece2);
					penum->add(piece3);
					penum->add(bking);
					penum->init();
					pegbb[side]->open(path,state2);
				}
			}
			check_presence(pegbb);
		}
	}
	for(piece1 = wqueen; piece1 <= wpawn; piece1++) {
		for(piece2 = piece1; piece2 <= wpawn; piece2++) {
			for(piece3 = bqueen; piece3 <= bpawn; piece3++) {
				for(side = white; side <= black;side++) {
					tab_index = EGBB::GetIndex(side,piece1,piece2,piece3);
			        pegbb[side] = egbbs[tab_index] = new EGBB;
					penum = &pegbb[side]->enumerator;
					penum->player = side;
					penum->add(wking);
					penum->add(piece1);
					penum->add(piece2);
					penum->add(bking);
					penum->add(piece3);
					penum->init();
					pegbb[side]->open(path,state2);
				}
				check_presence(pegbb);
			}
		}
	}

	printf("\rEgbbs loaded !      \n");
	fflush(stdout);
}

/*
Exported functions : probe_egbb & load_egbb
*/

int SEARCHER::get_score(
               int alpha,int beta,
			   int side, int w_king, int b_king,
			   int piece1, int square1,
			   int piece2, int square2,
			   int piece3, int square3
			   ) {

	int score;
	int move,from,to;
	int sq1,sq2,sq3,legal_moves;

	UBMP32 pos_index,tab_index;
	EGBB* pegbb;

	if(piece1 == 0)
		return DRAW;

	get_index(pos_index,tab_index,
		      side,w_king,b_king, 
		      piece1,square1,
			  piece2,square2,
			  piece3,square3);

	pegbb = egbbs[tab_index];


	if(pegbb->is_loaded) {
		score = pegbb->get_score(pos_index,this);
		return score;
	} else if(pegbb->use_search) {
		if(piece3) {
			square1 = SQ6488(square1);
			square2 = SQ6488(square2);
			square3 = SQ6488(square3);
		} else if(piece2) {
			square1 = SQ6488(square1);
			square2 = SQ6488(square2);
		} else {
			square1 = SQ6488(square1);
		}

		//set up position if we are at the root
		if(ply == 0) {
			set_pos(side,SQ6488(w_king),SQ6488(b_king),
				piece1,square1,
				piece2,square2,
				piece3,square3);
		}

		//generate moves and search
		pstack->count = 0;
		gen_caps();
		gen_noncaps();

		legal_moves = 0;

		for(int i = 0;i < pstack->count; i++) {
			
			move = pstack->move_st[i];
			PUSH_MOVE(move);
			if(attacks(player,plist[COMBINE(opponent,king)]->sq)) {
				POP_MOVE(move);
				continue;
			}

			legal_moves++;

			from = m_from(move);
			to = m_to(move);

			//squares
			sq1 = square1;
			sq2 = square2;
			sq3 = square3;

			//remove captured piece
			if(m_capture(move)) {
				if(sq1 == to) {     
					sq1 = sq2;
					sq2 = sq3;
					sq3 = INVALID;
				} else if(sq2 == to) {
					sq2 = sq3;
					sq3 = INVALID;
				} else if(sq3 == to) {
					sq3 = INVALID;
				}
			}
			//move piece
			if(sq1 == from)     
				sq1 = to;
			else if(sq2 == from)
				sq2 = to;
			else if(sq3 == from)
				sq3 = to;
			
			//recursive call
			if(sq1 == INVALID) {
				score = -DRAW;
			} else if(sq2 == INVALID) {
				score = -get_score(-beta,-alpha,
					player,SQ8864(plist[wking]->sq),SQ8864(plist[bking]->sq),
					board[sq1],SQ8864(sq1));
			} else if(sq3 == INVALID) {
				score = -get_score(-beta,-alpha,
					player,SQ8864(plist[wking]->sq),SQ8864(plist[bking]->sq),
					board[sq1],SQ8864(sq1), 
					board[sq2],SQ8864(sq2));
			} else {
				score = -get_score(-beta,-alpha,
					player,SQ8864(plist[wking]->sq),SQ8864(plist[bking]->sq),
					board[sq1],SQ8864(sq1), 
					board[sq2],SQ8864(sq2),
					board[sq3],SQ8864(sq3));
			}

			POP_MOVE(move);

			//update alpha
			if(score > alpha) { 
				alpha = score;
				if(score >= beta) {
					return beta;
				}
			}
		}

		//mate/stalemate?
		if(legal_moves == 0) {
			if(attacks(opponent,plist[COMBINE(player,king)]->sq))
				return LOSS;
			else
				return DRAW;
		}

		return alpha;
	} else {
        return DONT_KNOW;
	}
}
/*
King square tables
*/
static const int piece_cv[14] = {0,0,9,5,3,3,1,0,-9,-5,-3,-3,-1,0};
static const int  king_pcsq[64] = {
	-3,-2,-1,  0,  0,-1,-2,-3,  
	-2,-1,  0, 1, 1,  0,-1,-2,  
	-1,  0, 1, 2, 2, 1,  0,-1,  
	  0, 1, 2, 3, 3, 2, 1,  0,  
	  0, 1, 2, 3, 3, 2, 1,  0,  
	-1,  0, 1, 2, 2, 1,  0,-1,  
	-2,-1,  0, 1, 1,  0,-1,-2,  
	-3,-2,-1,  0,  0,-1,-2,-3
};

static const int  kbnk_pcsq[64] = {
	 7, 6, 5, 4, 3, 2, 1, 0,
     6, 7, 6, 5, 4, 3, 2, 1,  
     5, 6, 7, 6, 5, 4, 3, 2,  
     4, 5, 6, 7, 6, 5, 4, 3,  
     3, 4, 5, 6, 7, 6, 5, 4,  
     2, 3, 4, 5, 6, 7, 6, 5,  
	 1, 2, 3, 4, 5, 6, 7, 6,  
	 0, 1, 2, 3, 4, 5, 6, 7  
};

int probe_egbb_xxx(int player, int w_king, int b_king,
			   int piece1, int square1,
			   int piece2, int square2,
			   int piece3, int square3
			   ) {

    
	register int wdl_score,score,all_c,p_dist,material,ktable;

	//get score
	if(piece1 == _EMPTY)
		square1 = INVALID;
	if(piece2 == _EMPTY)
		square2 = INVALID;
	if(piece3 == _EMPTY)
		square3 = INVALID;

	PSEARCHER psearcher;
	l_lock(searcher_lock);
	for(int i = 0;i < 8;i++) {
		if(!searchers[i].used) {
			psearcher = &searchers[i];
			psearcher->used = 1;
			break;
		}
	}
	l_unlock(searcher_lock);
	
	wdl_score = psearcher->get_score(
		      LOSS,WIN,
		      player,w_king,b_king, 
		      piece1,square1,
			  piece2,square2,
			  piece3,square3);

    psearcher->used = 0;

	if(wdl_score == DONT_KNOW) {
		return _NOTFOUND;
	} else if(wdl_score == DRAW) {
		return 0;
	} else {

        /* 
		modify score to ensure progress
		*/

		//material
		material = piece_cv[piece1] + piece_cv[piece2] + piece_cv[piece3];
		if(player == black) material = -material;

		//passed pawn
		int prom_score,wp_prom = 0,bp_prom = 0;
		if(PIECE(piece1) == pawn) {
			if(COLOR(piece1) == white) wp_prom += rank64(square1);
			else bp_prom += (RANK7 - rank64(square1));
		}
		if(PIECE(piece2) == pawn) {
			if(COLOR(piece2) == white) wp_prom += rank64(square2);
			else bp_prom += (RANK7 - rank64(square2));
		}
		if(PIECE(piece3) == pawn) {
			if(COLOR(piece3) == white) wp_prom += rank64(square3);
			else bp_prom += (RANK7 - rank64(square3));
		}
		prom_score = wp_prom - bp_prom;
		if(player == black) prom_score = -prom_score;


		//king closeness to pawns
		if(piece3) {
			all_c = 5;
			int wp_dist = 0,bp_dist = 0;
			if(PIECE(piece1) == pawn) {
				wp_dist += distance(SQ6488(square1),SQ6488(w_king));
				bp_dist += distance(SQ6488(square1),SQ6488(b_king));
				if(COLOR(piece1) == white) wp_prom += rank64(square1);
				else bp_prom += (RANK7 - rank64(square1));
			}
			
			if(PIECE(piece2) == pawn) {
				wp_dist += distance(SQ6488(square2),SQ6488(w_king));
				bp_dist += distance(SQ6488(square2),SQ6488(b_king));

			}
			
			if(PIECE(piece3) == pawn) {
				wp_dist += distance(SQ6488(square3),SQ6488(w_king));
				bp_dist += distance(SQ6488(square3),SQ6488(b_king));
			}
			
			if(player == white) p_dist = 2 * wp_dist - bp_dist;
			else p_dist = 2 * bp_dist - wp_dist;
			
		} else if(piece2) {
			all_c = 4;
			p_dist = 0;
		} else {
			all_c = 3;
			p_dist = 0;
		}

		//king square table for kbnk and others
		if(all_c == 4 
			&& (((piece1 == wbishop && piece2 == wknight) ||
			     (piece1 == wknight && piece2 == wbishop))  
				 ||
				((piece1 == bbishop && piece2 == bknight) ||
			     (piece1 == bknight && piece2 == bbishop)))
			) {

			//get the king's & bishop's square
            int b_sq,loser_ksq,winner_ksq;
			if(COLOR(piece1) == white) {
				winner_ksq = w_king;
				loser_ksq = b_king;
				if(piece1 == wbishop) {
					b_sq = square1;
				} else {
					b_sq = square2;
				}
			} else {
				winner_ksq = b_king;
				loser_ksq = w_king;
				if(piece1 == bbishop) {
					b_sq = square1;
				} else {
					b_sq = square2;
				}
			}

			if(is_light64(b_sq));
			else {
				winner_ksq = MIRRORF64(winner_ksq); 
				loser_ksq = MIRRORF64(loser_ksq);
			}

            //score
			if(wdl_score == WIN) {
				ktable = king_pcsq[winner_ksq] - kbnk_pcsq[loser_ksq];
			} else {
				ktable = -king_pcsq[winner_ksq] + kbnk_pcsq[loser_ksq];
			}

		} else {
			if(player == white) {
				if(wdl_score == WIN) {
					ktable = - 2 * king_pcsq[b_king] + king_pcsq[w_king];
				} else {
					ktable = + 2 * king_pcsq[w_king] - king_pcsq[b_king];
				}
			} else {
				if(wdl_score == WIN) {
					ktable = - 2 * king_pcsq[w_king] + king_pcsq[b_king];
				} else {
					ktable = + 2 * king_pcsq[b_king] - king_pcsq[w_king];
				}
			}
		}

		//combine score
		if(player == white) {
			if(wdl_score == WIN) {
				score = 
					+ WIN_SCORE
					+ (5 - all_c) * 200 
					- distance(SQ6488(w_king),SQ6488(b_king)) * 2
                    + ktable * 5
					+ material * 50
					+ prom_score * 20
					- p_dist * 7;
			} else {
				score = 
					- WIN_SCORE
					- (5 - all_c) * 200
					+ distance(SQ6488(w_king),SQ6488(b_king)) * 2
					+ ktable * 5
					+ material * 50
					+ prom_score * 20
					- p_dist * 7;
			}
		} else {
			if(wdl_score == WIN) {
				score = 
					+ WIN_SCORE
					+ (5 - all_c) * 200
					- distance(SQ6488(w_king),SQ6488(b_king)) * 2
					+ ktable * 5
					+ material * 50
					+ prom_score * 20
					- p_dist * 7;
			} else {
				score = 
					- WIN_SCORE
					- (5 - all_c) * 200
					+ distance(SQ6488(w_king),SQ6488(b_king)) * 2
					+ ktable * 5
					+ material * 50
					+ prom_score * 20
					- p_dist * 7;
			}
		}
		return score;
	}
}

/*
old 
*/
DLLExport void CDECL load_egbb(char* path) {
	load_egbb_xxx(path,4194304,LOAD_4MEN);
}
DLLExport int CDECL probe_egbb(int player, int w_king, int b_king,
			    int piece1, int square1,
			    int piece2, int square2
			   ) {
	return probe_egbb_xxx(player,w_king,b_king,
		                  piece1,square1,
						  piece2,square2,
						  _EMPTY,0);
}
/*
new
*/
DLLExport void CDECL load_egbb_5men(char* path,int cache_size,int load_options) {
        load_egbb_xxx(path,cache_size,load_options);
}
DLLExport int  CDECL probe_egbb_5men(int player, int w_king, int b_king,
			    int piece1, int square1,
			    int piece2, int square2,
				int piece3, int square3
			   ) {
	return probe_egbb_xxx(player,w_king,b_king,
		                  piece1,square1,
						  piece2,square2,
						  piece3,square3);
}
